package com.ibm.oti.shared;

import java.lang.ref.WeakReference;
import java.security.AccessControlException;
import java.util.HashMap;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2010  All Rights Reserved
 */

/**
 * Abstract SharedHelper factory provides common functions and data to factory subclasses.
 * <p>
 * @author		OTI
 * @version		initial
 *
 * @see SharedClassHelperFactoryImpl
 * @see SharedDataHelperFactoryImpl
 */
public abstract class SharedAbstractHelperFactory {

	private static int idCount;		/* Single id count shared by all factories */
	static Object monitor;

	private static class Monitor {}

	static {
		idCount = 1;			/* bootstrap loader is 0 */
		monitor = new Monitor();		/* anything which changes idCount must synchronize on the monitor */
	}

	/* No guarantee that loader.toString() is unique, therefore need linked
	 * list implementation for clashing hashtable keys */
	class HelperList {
		private WeakReference helperRef, nextRef;
		public HelperList(SharedHelper current) {
			this.helperRef = new WeakReference(current);
			this.nextRef = null;
		}
		SharedHelper getHelper() {
			if (helperRef != null) {
				return (SharedHelper)helperRef.get();
			}
			return null;
		}
		void setHelper(SharedHelper helper) {
			if (helper != null) {
				helperRef = new WeakReference(helper);
			} else {
				helperRef = null;
			}
		}
		HelperList getNext() {
			if (nextRef != null) {
				return (HelperList)nextRef.get();
			}
			return null;
		}
		void setNext(HelperList list) {
			if (list != null) {
				nextRef = new WeakReference(list);
			} else {
				nextRef = null;
			}
		}
	}

	boolean checkPermission(ClassLoader loader, String type) {
		boolean result = true;
		SecurityManager sm = System.getSecurityManager();
		if (sm!=null) {
			try {
				sm.checkPermission(new SharedClassPermission(loader, type));
			} catch (AccessControlException e) {
				result = false;
			}
		}
		return result;
	}

	boolean canFind(ClassLoader loader) {
		return checkPermission(loader, "read");
	}

	boolean canStore(ClassLoader loader) {
		return checkPermission(loader, "write");
	}

	/* CMVC 100176: toString() can be over-ridden in ClassLoaders to return null */
	String objectToString(Object o) {
		return o.getClass().getName() + Integer.toHexString(o.hashCode());
	}

	/* Should be always called from synchronized block on HashMap returned by getHelperMap */
	void addHelper(ClassLoader loader, SharedHelper helper) {
		String loaderString = objectToString(loader);
		HashMap helpers = getHelperMap();
		WeakReference listRef = (WeakReference)helpers.get(loaderString);
		HelperList listHead = (listRef != null) ? (HelperList)listRef.get() : null;
		HelperList list = listHead;

		if (list==null) {
			helpers.put(loaderString, new WeakReference(new HelperList(helper)));
		} else {
			HelperList tail = null;

			while (list != null) {
				if (list.getHelper() == null) {
					list.setHelper(helper);
					return;
				}
				tail = list;
				list = list.getNext();
			}
			tail.setNext(new HelperList(helper));
		}
	}

	SharedHelper findSharedHelperForClassLoader(ClassLoader loader) {
		if (loader==null) {
			return null;
		}
		HashMap helpers = getHelperMap();
		synchronized(helpers) {
			String loaderString = objectToString(loader);
			WeakReference listRef = (WeakReference)helpers.get(loaderString);
			HelperList listHead = (listRef != null) ? (HelperList)listRef.get() : null;
			HelperList list = listHead;

			while (list != null) {
				SharedHelper helper = list.getHelper();

				if (helper != null) {
					ClassLoader currentLoader = helper.getClassLoader();

					if (currentLoader == loader) {
						return helper;
					} else if (currentLoader == null) {
						/* ClassLoader has been Garbage Collected.
						 * Let go of the SharedClassHelper and re-use the HelperList */
						list.setHelper(null);
					}
				}
				list = list.getNext();
			}
		}
		return null;
	}

	/* Should return the HashMap which holds the SharedHelpers for this factory */
	abstract HashMap getHelperMap();

	int getNewID() {
		synchronized(monitor) {
			return idCount++;
		}
	}
}
